from __future__ import division  # confidence high
from __future__ import with_statement

import gzip
import os
import warnings
import zipfile

import numpy as np

import pyfits
from pyfits.convenience import _getext
from pyfits.util import BytesIO
from pyfits.tests import PyfitsTestCase
from pyfits.tests.util import catch_warnings, ignore_warnings, CaptureStdio

from nose.tools import assert_equal, assert_raises, assert_true, assert_false


class TestCore(PyfitsTestCase):
    def test_with_statement(self):
        with pyfits.open(self.data('ascii.fits')) as f:
            pass

    def test_missing_file(self):
        assert_raises(IOError, pyfits.open, self.temp('does-not-exist.fits'))

    def test_naxisj_check(self):
        hdulist = pyfits.open(self.data('o4sp040b0_raw.fits'))

        hdulist[1].header['NAXIS3'] = 500

        assert 'NAXIS3' in hdulist[1].header
        hdulist.verify('silentfix')
        assert 'NAXIS3' not in hdulist[1].header

    def test_byteswap(self):
        p = pyfits.PrimaryHDU()
        l = pyfits.HDUList()

        n = np.zeros(3, dtype='i2')
        n[0] = 1
        n[1] = 60000
        n[2] = 2

        c = pyfits.Column(name='foo', format='i2', bscale=1, bzero=32768,
                          array=n)
        t = pyfits.new_table([c])

        l.append(p)
        l.append(t)

        l.writeto(self.temp('test.fits'), clobber=True)

        with pyfits.open(self.temp('test.fits')) as p:
            assert_equal(p[1].data[1]['foo'], 60000.0)

    def test_add_del_columns(self):
        p = pyfits.ColDefs([])
        p.add_col(pyfits.Column(name='FOO', format='3J'))
        p.add_col(pyfits.Column(name='BAR', format='1I'))
        assert_equal(p.names, ['FOO', 'BAR'])
        p.del_col('FOO')
        assert_equal(p.names, ['BAR'])

    def test_add_del_columns2(self):
        hdulist = pyfits.open(self.data('tb.fits'))
        table = hdulist[1]
        assert_equal(table.data.dtype.names, ('c1', 'c2', 'c3', 'c4'))
        assert_equal(table.columns.names, ['c1', 'c2', 'c3', 'c4'])
        #old_data = table.data.base.copy().view(pyfits.FITS_rec)
        table.columns.del_col('c1')
        assert_equal(table.data.dtype.names, ('c2', 'c3', 'c4'))
        assert_equal(table.columns.names, ['c2', 'c3', 'c4'])

        #for idx in range(len(old_data)):
        #    assert_equal(np.all(old_data[idx][1:], table.data[idx]))

        table.columns.del_col('c3')
        assert_equal(table.data.dtype.names, ('c2', 'c4'))
        assert_equal(table.columns.names, ['c2', 'c4'])

        #for idx in range(len(old_data)):
        #    assert_equal(np.all(old_data[idx][2:], table.data[idx]))

        table.columns.add_col(pyfits.Column('foo', '3J'))
        assert_equal(table.data.dtype.names, ('c2', 'c4', 'foo'))
        assert_equal(table.columns.names, ['c2', 'c4', 'foo'])

        hdulist.writeto(self.temp('test.fits'), clobber=True)
        with pyfits.open(self.temp('test.fits')) as hdulist:
            table = hdulist[1]
            assert_equal(table.data.dtype.names, ('c1', 'c2', 'c3'))
            assert_equal(table.columns.names, ['c1', 'c2', 'c3'])

    def test_update_header_card(self):
        """A very basic test for the Header.update method--I'd like to add a
        few more cases to this at some point.
        """

        header = pyfits.Header()
        comment = 'number of bits per data pixel'
        header['BITPIX'] = (16, comment)
        assert 'BITPIX' in header
        assert_equal(header['BITPIX'], 16)
        assert_equal(header.ascard['BITPIX'].comment, comment)

        # The new API doesn't support savecomment so leave this line here; at
        # any rate good to have testing of the new API mixed with the old API
        header.update('BITPIX', 32, savecomment=True)
        # Make sure the value has been updated, but the comment was preserved
        assert_equal(header['BITPIX'], 32)
        assert_equal(header.ascard['BITPIX'].comment, comment)

        # The comment should still be preserved--savecomment only takes effect if
        # a new comment is also specified
        header['BITPIX'] = 16
        assert_equal(header.ascard['BITPIX'].comment, comment)
        header.update('BITPIX', 16, 'foobarbaz', savecomment=True)
        assert_equal(header.ascard['BITPIX'].comment, comment)

    def test_set_card_value(self):
        """Similar to test_update_header_card(), but tests the the
        `header['FOO'] = 'bar'` method of updating card values.
        """

        header = pyfits.Header()
        comment = 'number of bits per data pixel'
        card = pyfits.Card.fromstring('BITPIX  = 32 / %s' % comment)
        header.ascard.append(card)

        header['BITPIX'] = 32

        assert 'BITPIX' in header
        assert_equal(header['BITPIX'], 32)
        assert_equal(header.ascard['BITPIX'].key, 'BITPIX')
        assert_equal(header.ascard['BITPIX'].value, 32)
        assert_equal(header.ascard['BITPIX'].comment, comment)

    def test_uint(self):
        hdulist_f = pyfits.open(self.data('o4sp040b0_raw.fits'))
        hdulist_i = pyfits.open(self.data('o4sp040b0_raw.fits'), uint=True)

        assert_equal(hdulist_f[1].data.dtype, np.float32)
        assert_equal(hdulist_i[1].data.dtype, np.uint16)
        assert np.all(hdulist_f[1].data == hdulist_i[1].data)

    def test_fix_missing_card_append(self):
        hdu = pyfits.ImageHDU()
        errs = hdu.req_cards('TESTKW', None, None, 'foo', 'silentfix', [])
        assert_equal(len(errs), 1)
        assert 'TESTKW' in hdu.header
        assert_equal(hdu.header['TESTKW'], 'foo')
        assert_equal(hdu.header.ascard[-1].key, 'TESTKW')

    def test_fix_invalid_keyword_value(self):
        hdu = pyfits.ImageHDU()
        hdu.header['TESTKW'] = 'foo'
        errs = hdu.req_cards('TESTKW', None,
                             lambda v: v == 'foo', 'foo', 'ignore', [])
        assert_equal(len(errs), 0)

        # Now try a test that will fail, and ensure that an error will be
        # raised in 'exception' mode
        errs = hdu.req_cards('TESTKW', None, lambda v: v == 'bar', 'bar',
                             'exception', [])
        assert_equal(len(errs), 1)
        assert_equal(errs[0], "'TESTKW' card has invalid value 'foo'.")

        # See if fixing will work
        hdu.req_cards('TESTKW', None, lambda v: v == 'bar', 'bar', 'silentfix',
                      [])
        assert_equal(hdu.header['TESTKW'], 'bar')

    def test_unfixable_missing_card(self):
        class TestHDU(pyfits.hdu.base.NonstandardExtHDU):
            def _verify(self, option='warn'):
                errs = super(TestHDU, self)._verify(option)
                hdu.req_cards('TESTKW', None, None, None, 'fix', errs)
                return errs

        hdu = TestHDU(header=pyfits.Header())
        assert_raises(pyfits.VerifyError, hdu.verify, 'fix')

    def test_exception_on_verification_error(self):
        hdu = pyfits.ImageHDU()
        del hdu.header['XTENSION']
        assert_raises(pyfits.VerifyError, hdu.verify, 'exception')

    def test_ignore_verification_error(self):
        hdu = pyfits.ImageHDU()
        # The default here would be to issue a warning; ensure that no warnings
        # or exceptions are raised
        with catch_warnings():
            warnings.simplefilter('error')
            del hdu.header['NAXIS']
            try:
                hdu.verify('ignore')
            except Exception, e:
                self.fail('An exception occurred when the verification error '
                          'should have been ignored: %s' % e)
        # Make sure the error wasn't fixed either, silently or otherwise
        assert_false('NAXIS' in hdu.header)

    def test_unrecognized_verify_option(self):
        hdu = pyfits.ImageHDU()
        assert_raises(ValueError, hdu.verify, 'foobarbaz')

    def test_getext(self):
        """
        Test the various different ways of specifying an extension header in
        the convenience functions.
        """

        hl, ext = _getext(self.data('test0.fits'), 'readonly', 1)
        assert_equal(ext, 1)
        assert_raises(ValueError, _getext, self.data('test0.fits'), 'readonly',
                      1, 2)
        assert_raises(ValueError, _getext, self.data('test0.fits'), 'readonly',
                      (1, 2))
        assert_raises(ValueError, _getext, self.data('test0.fits'), 'readonly',
                      'sci', 'sci')
        assert_raises(TypeError, _getext, self.data('test0.fits'), 'readonly',
                      1, 2, 3)
        hl, ext = _getext(self.data('test0.fits'), 'readonly', ext=1)
        assert_equal(ext, 1)
        hl, ext = _getext(self.data('test0.fits'), 'readonly', ext=('sci', 2))
        assert_equal(ext, ('sci', 2))
        assert_raises(TypeError, _getext, self.data('test0.fits'), 'readonly',
                      1, ext=('sci', 2), extver=3)
        assert_raises(TypeError, _getext, self.data('test0.fits'), 'readonly',
                      ext=('sci', 2), extver=3)

        hl, ext = _getext(self.data('test0.fits'), 'readonly', 'sci')
        assert_equal(ext, ('sci', 1))
        hl, ext = _getext(self.data('test0.fits'), 'readonly', 'sci', 1)
        assert_equal(ext, ('sci', 1))
        hl, ext = _getext(self.data('test0.fits'), 'readonly', ('sci', 1))
        assert_equal(ext, ('sci', 1))
        hl, ext = _getext(self.data('test0.fits'), 'readonly', 'sci',
                          extver=1, do_not_scale_image_data=True)
        assert_equal(ext, ('sci', 1))
        assert_raises(TypeError, _getext, self.data('test0.fits'), 'readonly',
                      'sci', ext=1)
        assert_raises(TypeError, _getext, self.data('test0.fits'), 'readonly',
                      'sci', 1, extver=2)

        hl, ext = _getext(self.data('test0.fits'), 'readonly', extname='sci')
        assert_equal(ext, ('sci', 1))
        hl, ext = _getext(self.data('test0.fits'), 'readonly', extname='sci',
                          extver=1)
        assert_equal(ext, ('sci', 1))
        assert_raises(TypeError, _getext, self.data('test0.fits'), 'readonly',
                      extver=1)

    def test_extension_name_case_sensitive(self):
        """
        Tests that setting pyfits.EXTENSION_NAME_CASE_SENSITIVE at runtime
        works.
        """

        if 'PYFITS_EXTENSION_NAME_CASE_SENSITIVE' in os.environ:
            del os.environ['PYFITS_EXTENSION_NAME_CASE_SENSITIVE']

        hdu = pyfits.ImageHDU()
        hdu.name = 'sCi'
        assert_equal(hdu.name, 'SCI')
        assert_equal(hdu.header['EXTNAME'], 'SCI')

        try:
            pyfits.EXTENSION_NAME_CASE_SENSITIVE = True
            hdu = pyfits.ImageHDU()
            hdu.name = 'sCi'
            assert_equal(hdu.name, 'sCi')
            assert_equal(hdu.header['EXTNAME'], 'sCi')
        finally:
            pyfits.EXTENSION_NAME_CASE_SENSITIVE = False

        hdu.name = 'sCi'
        assert_equal(hdu.name, 'SCI')
        assert_equal(hdu.header['EXTNAME'], 'SCI')

    def test_hdu_fromstring(self):
        """
        Tests creating a fully-formed HDU object from a string containing the
        bytes of the HDU.
        """

        dat = open(self.data('test0.fits'), 'rb').read()

        offset = 0
        with pyfits.open(self.data('test0.fits')) as hdul:
            hdulen = hdul[0]._datLoc + hdul[0]._datSpan
            hdu = pyfits.PrimaryHDU.fromstring(dat[:hdulen])
            assert_true(isinstance(hdu, pyfits.PrimaryHDU))
            assert_equal(hdul[0].header, hdu.header)
            assert_true(hdu.data is None)

        hdu.header['TEST'] = 'TEST'
        hdu.writeto(self.temp('test.fits'))
        with pyfits.open(self.temp('test.fits')) as hdul:
            assert_true(isinstance(hdu, pyfits.PrimaryHDU))
            assert_equal(hdul[0].header[:-1], hdu.header[:-1])
            assert_equal(hdul[0].header['TEST'], 'TEST')
            assert_true(hdu.data is None)

        with pyfits.open(self.data('test0.fits'))as hdul:
            for ext_hdu in hdul[1:]:
                offset += hdulen
                hdulen = len(str(ext_hdu.header)) + ext_hdu._datSpan
                hdu = pyfits.ImageHDU.fromstring(dat[offset:offset + hdulen])
                assert_true(isinstance(hdu, pyfits.ImageHDU))
                assert_equal(ext_hdu.header, hdu.header)
                assert_true((ext_hdu.data == hdu.data).all())

    def test_nonstandard_hdu(self):
        """
        Regression test for #157.  Tests that "Nonstandard" HDUs with SIMPLE =
        F are read and written without prepending a superfluous and unwanted
        standard primary HDU.
        """

        data = np.arange(100, dtype=np.uint8)
        hdu = pyfits.PrimaryHDU(data=data)
        hdu.header['SIMPLE'] = False
        hdu.writeto(self.temp('test.fits'))

        info = [(0, '', 'NonstandardHDU', 5, (), '', '')]
        with pyfits.open(self.temp('test.fits')) as hdul:
            assert_equal(hdul.info(output=False), info)
            # NonstandardHDUs just treat the data as an unspecified array of
            # bytes.  The first 100 bytes should match the original data we
            # passed in...the rest should be zeros padding out the rest of the
            # FITS block
            assert_true((hdul[0].data[:100] == data).all())
            assert_true((hdul[0].data[100:] == 0).all())


class TestConvenienceFunctions(PyfitsTestCase):
    def test_writeto(self):
        """
        Simple test for writing a trivial header and some data to a file
        with the `writeto()` convenience function.
        """

        data = np.zeros((100, 100))
        header = pyfits.Header()
        pyfits.writeto(self.temp('array.fits'), data, header=header,
                       clobber=True)
        hdul = pyfits.open(self.temp('array.fits'))
        assert_equal(len(hdul), 1)
        assert_true((data == hdul[0].data).all())

    def test_writeto_2(self):
        """
        Test of `writeto()` with a trivial header containing a single keyword;
        regression for #107.
        """

        data = np.zeros((100,100))
        header = pyfits.Header()
        header.update('CRPIX1', 1.)
        pyfits.writeto(self.temp('array.fits'), data, header=header,
                       clobber=True, output_verify='silentfix')
        hdul = pyfits.open(self.temp('array.fits'))
        assert_equal(len(hdul), 1)
        assert_true((data == hdul[0].data).all())
        assert_true('CRPIX1' in hdul[0].header)
        assert_equal(hdul[0].header['CRPIX1'], 1.0)


class TestFileFunctions(PyfitsTestCase):
    """
    Tests various basic I/O operations, specifically in the pyfits.file._File
    class.
    """

    def test_open_gzipped(self):
        with ignore_warnings():
            assert_equal(len(pyfits.open(self._make_gzip_file())), 5)

    def test_detect_gzipped(self):
        """Test detection of a gzip file when the extension is not .gz."""

        with ignore_warnings():
            assert_equal(len(pyfits.open(self._make_gzip_file('test0.fz'))), 5)

    def test_open_zipped(self):
        with ignore_warnings():
            assert_equal(len(pyfits.open(self._make_zip_file())), 5)

    def test_detect_zipped(self):
        """Test detection of a zip file when the extension is not .zip."""

        zf = self._make_zip_file(filename='test0.fz')
        with ignore_warnings():
            assert_equal(len(pyfits.open(zf)), 5)

    def test_open_zipped_writeable(self):
        """Opening zipped files in a writeable mode should fail."""

        zf = self._make_zip_file()
        assert_raises(IOError, pyfits.open, zf, 'update')
        assert_raises(IOError, pyfits.open, zf, 'append')

    def test_open_multiple_member_zipfile(self):
        """
        Opening zip files containing more than one member files should fail
        as there's no obvious way to specify which file is the FITS file to
        read.
        """

        zfile = zipfile.ZipFile(self.temp('test0.zip'), 'w')
        zfile.write(self.data('test0.fits'))
        zfile.writestr('foo', 'bar')
        zfile.close()

        assert_raises(IOError, pyfits.open, zfile.filename)

    def test_read_open_file(self):
        """Read from an existing file object."""

        with open(self.data('test0.fits'), 'rb') as f:
            assert_equal(len(pyfits.open(f)), 5)

    def test_read_closed_file(self):
        """Read from an existing file object that's been closed."""

        f = open(self.data('test0.fits'), 'rb')
        f.close()
        assert_equal(len(pyfits.open(f)), 5)

    def test_read_open_gzip_file(self):
        """Read from an open gzip file object."""

        gf = gzip.GzipFile(self._make_gzip_file())
        try:
            assert_equal(len(pyfits.open(gf)), 5)
        finally:
            gf.close()

    def test_open_gzip_file_for_writing(self):
        """Regression test for #195."""

        gf = self._make_gzip_file()
        with pyfits.open(gf, mode='update') as h:
            h[0].header['EXPFLAG'] = 'ABNORMAL'
        with pyfits.open(gf) as h:
            # Just to make sur ethe update worked; if updates work
            # normal writes should work too...
            assert_equal(h[0].header['EXPFLAG'], 'ABNORMAL')

    def test_read_file_like_object(self):
        """Test reading a FITS file from a file-like object."""

        filelike = BytesIO()
        with open(self.data('test0.fits'), 'rb') as f:
            filelike.write(f.read())
        filelike.seek(0)
        with ignore_warnings():
            assert_equal(len(pyfits.open(filelike)), 5)

    def test_updated_file_permissions(self):
        """
        Regression test for #79.  Tests that when a FITS file is modified in
        update mode, the file permissions are preserved.
        """

        filename = self.temp('test.fits')
        hdul = [pyfits.PrimaryHDU(), pyfits.ImageHDU()]
        hdul = pyfits.HDUList(hdul)
        hdul.writeto(filename)

        old_mode = os.stat(filename).st_mode

        hdul = pyfits.open(filename, mode='update')
        hdul.insert(1, pyfits.ImageHDU())
        hdul.flush()
        hdul.close()

        assert_equal(old_mode, os.stat(filename).st_mode)

    def _make_gzip_file(self, filename='test0.fits.gz'):
        gzfile = self.temp(filename)
        with open(self.data('test0.fits'), 'rb') as f:
            gz = gzip.open(gzfile, 'wb')
            gz.write(f.read())
            gz.close()

        return gzfile

    def _make_zip_file(self, mode='copyonwrite', filename='test0.fits.zip'):
        zfile = zipfile.ZipFile(self.temp(filename), 'w')
        zfile.write(self.data('test0.fits'))
        zfile.close()

        return zfile.filename


class TestStreamingFunctions(PyfitsTestCase):
    """Test functionality of the StreamingHDU class."""

    def test_streaming_hdu(self):
        shdu = self._make_streaming_hdu(self.temp('new.fits'))
        assert_true(isinstance(shdu.size, int))
        assert_equal(shdu.size, 100)

    def test_streaming_hdu_file_wrong_mode(self):
        """
        Test that streaming an HDU to a file opened in the wrong mode fails as
        expected.
        """

        with open(self.temp('new.fits'), 'wb') as f:
            header = pyfits.Header()
            assert_raises(ValueError, pyfits.StreamingHDU, f, header)

    def test_streaming_hdu_write_file(self):
        """Test streaming an HDU to an open file object."""

        arr = np.zeros((5, 5), dtype=np.int32)
        with open(self.temp('new.fits'), 'ab+') as f:
            shdu = self._make_streaming_hdu(f)
            shdu.write(arr)
            assert_true(shdu.writecomplete)
            assert_equal(shdu.size, 100)
        hdul = pyfits.open(self.temp('new.fits'))
        assert_equal(len(hdul), 1)
        assert_true((hdul[0].data == arr).all())

    def test_streaming_hdu_write_file_like(self):
        """Test streaming an HDU to an open file-like object."""

        arr = np.zeros((5, 5), dtype=np.int32)
        # The file-like object underlying a StreamingHDU must be in binary mode
        sf = BytesIO()
        shdu = self._make_streaming_hdu(sf)
        shdu.write(arr)
        assert_true(shdu.writecomplete)
        assert_equal(shdu.size, 100)

        sf.seek(0)
        hdul = pyfits.open(sf)
        assert_equal(len(hdul), 1)
        assert_true((hdul[0].data == arr).all())

    def test_streaming_hdu_append_extension(self):
        arr = np.zeros((5, 5), dtype=np.int32)
        with open(self.temp('new.fits'), 'ab+') as f:
            shdu = self._make_streaming_hdu(f)
            shdu.write(arr)
        # Doing this again should update the file with an extension
        with open(self.temp('new.fits'), 'ab+') as f:
            shdu = self._make_streaming_hdu(f)
            shdu.write(arr)

    def test_fix_invalid_extname(self):
        phdu = pyfits.PrimaryHDU()
        ihdu = pyfits.ImageHDU()
        ihdu.header['EXTNAME'] = 12345678
        hdul = pyfits.HDUList([phdu, ihdu])

        assert_raises(pyfits.VerifyError, hdul.writeto, self.temp('temp.fits'),
                      output_verify='exception')
        with CaptureStdio():
            hdul.writeto(self.temp('temp.fits'), output_verify='fix')
        with pyfits.open(self.temp('temp.fits')):
            assert_equal(hdul[1].name, '12345678')
            assert_equal(hdul[1].header['EXTNAME'], '12345678')

    def _make_streaming_hdu(self, fileobj):
        hd = pyfits.Header()
        hd['SIMPLE'] = (True, 'conforms to FITS standard')
        hd['BITPIX'] = (32, 'array data type')
        hd['NAXIS'] = (2, 'number of array dimensions')
        hd['NAXIS1'] = 5
        hd['NAXIS2'] = 5
        hd['EXTEND'] = True
        return pyfits.StreamingHDU(fileobj, hd)



from pyfits.hdu.base import register_hdu, unregister_hdu, DELAYED
from pyfits.hdu.compressed import CompImageHDU
from pyfits.hdu.groups import GroupsHDU, GroupData, Group
from pyfits.hdu.hdulist import HDUList
from pyfits.hdu.image import PrimaryHDU, ImageHDU
from pyfits.hdu.nonstandard import FitsHDU
from pyfits.hdu.streaming import StreamingHDU
from pyfits.hdu.table import TableHDU, BinTableHDU

__all__ = ['HDUList', 'PrimaryHDU', 'ImageHDU', 'TableHDU', 'BinTableHDU',
           'GroupsHDU', 'GroupData', 'Group', 'CompImageHDU', 'FitsHDU',
           'StreamingHDU', 'register_hdu', 'unregister_hdu', 'DELAYED']



import gzip
import os
import shutil
import sys
import warnings

import numpy as np
from numpy import memmap as Memmap

import pyfits
from pyfits.card import Card
from pyfits.column import _FormatP
from pyfits.file import PYTHON_MODES, _File
from pyfits.hdu import compressed
from pyfits.hdu.base import _BaseHDU, _ValidHDU, _NonstandardHDU, ExtensionHDU
from pyfits.hdu.compressed import CompImageHDU
from pyfits.hdu.groups import GroupsHDU
from pyfits.hdu.image import PrimaryHDU, ImageHDU
from pyfits.hdu.table import _TableBaseHDU
from pyfits.util import (_is_int, _tmp_name, _pad_length, BLOCK_SIZE, isfile,
                         fileobj_name, fileobj_closed, fileobj_mode,
                         ignore_sigint, _get_array_mmap, indent)
from pyfits.verify import _Verify, _ErrList, VerifyError, VerifyWarning


def fitsopen(name, mode='readonly', memmap=None, save_backup=False, **kwargs):
    """Factory function to open a FITS file and return an `HDUList` object.

    Parameters
    ----------
    name : file path, file object or file-like object
        File to be opened.

    mode : str
        Open mode, 'readonly' (default), 'update', 'append', 'denywrite', or
        'ostream'.

        If `name` is a file object that is already opened, `mode` must
        match the mode the file was opened with, copyonwrite (rb),
        readonly (rb), update (rb+), append (ab+), ostream (w),
        denywrite (rb)).

    memmap : bool
        Is memory mapping to be used?

    save_backup : bool
        If the file was opened in update or append mode, this ensures that a
        backup of the original file is saved before any changes are flushed.
        The backup has the same name as the original file with ".bak" appended.
        If "file.bak" already exists then "file.bak.1" is used, and so on.

    kwargs : dict
        optional keyword arguments, possible values are:

        - **uint** : bool

            Interpret signed integer data where ``BZERO`` is the
            central value and ``BSCALE == 1`` as unsigned integer
            data.  For example, `int16` data with ``BZERO = 32768``
            and ``BSCALE = 1`` would be treated as `uint16` data.

            Note, for backward compatibility, the kwarg **uint16** may
            be used instead.  The kwarg was renamed when support was
            added for integers of any size.

        - **ignore_missing_end** : bool

            Do not issue an exception when opening a file that is
            missing an ``END`` card in the last header.

        - **checksum** : bool, str

            If `True`, verifies that both ``DATASUM`` and
            ``CHECKSUM`` card values (when present in the HDU header)
            match the header and data of all HDU's in the file.  Updates to a
            file that already has a checksum will preserve and update the
            existing checksums unless this argument is given a value of
            'remove', in which case the CHECKSUM and DATASUM values are not
            checked, and are removed when saving changes to the file.

        - **disable_image_compression** : bool

            If `True`, treates compressed image HDU's like normal
            binary table HDU's.

        - **do_not_scale_image_data** : bool

            If `True`, image data is not scaled using BSCALE/BZERO values
            when read.

        - **scale_back** : bool

            If `True`, when saving changes to a file that contained scaled
            image data, restore the data to the original type and reapply the
            original BSCALE/BZERO values.  This could lead to loss of accuracy
            if scaling back to integer values after performing floating point
            operations on the data.

    Returns
    -------
        hdulist : an `HDUList` object
            `HDUList` containing all of the header data units in the
            file.

    """

    if memmap is None:
        from pyfits import USE_MEMMAP
        memmap = USE_MEMMAP

    if 'uint16' in kwargs and 'uint' not in kwargs:
        kwargs['uint'] = kwargs['uint16']
        del kwargs['uint16']

    if not name:
        raise ValueError('Empty filename: %s' % repr(name))

    return HDUList.fromfile(name, mode, memmap, save_backup, **kwargs)


class HDUList(list, _Verify):
    """
    HDU list class.  This is the top-level FITS object.  When a FITS
    file is opened, a `HDUList` object is returned.
    """

    def __init__(self, hdus=[], file=None):
        """
        Construct a `HDUList` object.

        Parameters
        ----------
        hdus : sequence of HDU objects or single HDU, optional
            The HDU object(s) to comprise the `HDUList`.  Should be
            instances of `_BaseHDU`.

        file : file object, optional
            The opened physical file associated with the `HDUList`.
        """

        self.__file = file
        self._save_backup = False

        if hdus is None:
            hdus = []

        # can take one HDU, as well as a list of HDU's as input
        if isinstance(hdus, _ValidHDU):
            hdus = [hdus]
        elif not isinstance(hdus, (HDUList, list)):
            raise TypeError("Invalid input for HDUList.")

        for idx, hdu in enumerate(hdus):
            if not isinstance(hdu, _BaseHDU):
                raise TypeError(
                      "Element %d in the HDUList input is not an HDU." % idx)

        super(HDUList, self).__init__(hdus)

        self.update_extend()

    def __iter__(self):
        for idx in range(len(self)):
            yield self[idx]

    def __getitem__(self, key):
        """
        Get an HDU from the `HDUList`, indexed by number or name.
        """

        if isinstance(key, slice):
            hdus = super(HDUList, self).__getitem__(key)
            return HDUList(hdus)

        idx = self.index_of(key)
        return super(HDUList, self).__getitem__(idx)

    def __setitem__(self, key, hdu):
        """
        Set an HDU to the `HDUList`, indexed by number or name.
        """

        _key = self.index_of(key)
        if isinstance(hdu, (slice, list)):
            if _is_int(_key):
                raise ValueError('An element in the HDUList must be an HDU.')
            for item in hdu:
                if not isinstance(item, _BaseHDU):
                    raise ValueError('%s is not an HDU.' % item)
        else:
            if not isinstance(hdu, _BaseHDU):
                raise ValueError('%s is not an HDU.' % hdu)

        try:
            super(HDUList, self).__setitem__(_key, hdu)
        except IndexError:
            raise IndexError('Extension %s is out of bound or not found.'
                             % key)
        self._resize = True
        self._truncate = False

    def __delitem__(self, key):
        """
        Delete an HDU from the `HDUList`, indexed by number or name.
        """

        if isinstance(key, slice):
            end_index = len(self)
        else:
            key = self.index_of(key)
            end_index = len(self) - 1

        super(HDUList, self).__delitem__(key)

        if (key == end_index or key == -1 and not self._resize):
            self._truncate = True
        else:
            self._truncate = False
            self._resize = True

    def __getslice__(self, start, end):
        return self[slice(start, end)]

    def __delslice__(self, start, stop):
        """
        Delete a slice of HDUs from the `HDUList`, indexed by number only.
        """

        del self[slice(start, stop)]

    # Support the 'with' statement
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    @classmethod
    def fromfile(cls, fileobj, mode='readonly', memmap=False,
                 save_backup=False, **kwargs):
        """
        Creates an HDUList instance from a file-like object.

        The actual implementation of :func:`fitsopen`, and generally shouldn't
        be used directly.  Use :func:`pyfits.open` instead (and see its
        documentation for details of the parameters accepted by this method).
        """

        return cls._readfrom(fileobj=fileobj, mode=mode, memmap=memmap,
                             save_backup=save_backup, **kwargs)

    @classmethod
    def fromstring(cls, data, **kwargs):
        """
        Creates an HDUList instance from a string or other in-memory data
        buffer containing an entire FITS file.  Similar to
        :meth:`HDUList.fromfile`, but does not accept the mode or memmap
        arguments, as they are only relevant to reading from a file on disk.

        This is useful for interfacing with other libraries such as CFITSIO,
        and may also be useful for streaming applications.

        Parameters
        ----------
        data : str, buffer, memoryview, etc.
            A string or other memory buffer containing an entire FITS file.  It
            should be noted that if that memory is read-only (such as a Python
            string) the returned :class:`HDUList`'s data portions will also be
            read-only.

        kwargs : dict
            Optional keyword arguments.  See :func:`pyfits.open` for details.

        Returns
        -------
        hdul : HDUList
            An :class:`HDUList` object representing the in-memory FITS file.
        """

        return cls._readfrom(data=data, **kwargs)

    def fileinfo(self, index):
        """
        Returns a dictionary detailing information about the locations
        of the indexed HDU within any associated file.  The values are
        only valid after a read or write of the associated file with
        no intervening changes to the `HDUList`.

        Parameters
        ----------
        index : int
            Index of HDU for which info is to be returned.

        Returns
        -------
        fileinfo : dict or None

            The dictionary details information about the locations of
            the indexed HDU within an associated file.  Returns `None`
            when the HDU is not associated with a file.

            Dictionary contents:

            ========== ========================================================
            Key        Value
            ========== ========================================================
            file       File object associated with the HDU
            filename   Name of associated file object
            filemode   Mode in which the file was opened (readonly,
                       copyonwrite, update, append, denywrite, ostream)
            resized    Flag that when `True` indicates that the data has been
                       resized since the last read/write so the returned values
                       may not be valid.
            hdrLoc     Starting byte location of header in file
            datLoc     Starting byte location of data block in file
            datSpan    Data size including padding
            ========== ========================================================

        """

        if self.__file is not None:
            output = self[index].fileinfo()

            if not output:
                # OK, the HDU associated with this index is not yet
                # tied to the file associated with the HDUList.  The only way
                # to get the file object is to check each of the HDU's in the
                # list until we find the one associated with the file.
                f = None

                for hdu in self:
                    info = hdu.fileinfo()

                    if info:
                        f = info['file']
                        fm = info['filemode']
                        break

                output = {'file': f, 'filemode': fm, 'hdrLoc': None,
                          'datLoc': None, 'datSpan': None}

            output['filename'] = self.__file.name
            output['resized'] = self._wasresized()
        else:
            output = None

        return output

    def insert(self, index, hdu):
        """
        Insert an HDU into the `HDUList` at the given `index`.

        Parameters
        ----------
        index : int
            Index before which to insert the new HDU.

        hdu : _BaseHDU instance
            The HDU object to insert
        """

        if not isinstance(hdu, _BaseHDU):
            raise ValueError('%s is not an HDU.' % hdu)

        num_hdus = len(self)

        if index == 0 or num_hdus == 0:
            if num_hdus != 0:
                # We are inserting a new Primary HDU so we need to
                # make the current Primary HDU into an extension HDU.
                if isinstance(self[0], GroupsHDU):
                    raise ValueError(
                        "The current Primary HDU is a GroupsHDU.  "
                        "It can't be made into an extension HDU, "
                        "so another HDU cannot be inserted before it.")

                hdu1 = ImageHDU(self[0].data, self[0].header)

                # Insert it into position 1, then delete HDU at position 0.
                super(HDUList, self).insert(1, hdu1)
                super(HDUList, self).__delitem__(0)

            if not isinstance(hdu, (PrimaryHDU, _NonstandardHDU)):
                # You passed in an Extension HDU but we need a Primary HDU.
                # If you provided an ImageHDU then we can convert it to
                # a primary HDU and use that.
                if isinstance(hdu, ImageHDU):
                    hdu = PrimaryHDU(hdu.data, hdu.header)
                else:
                    # You didn't provide an ImageHDU so we create a
                    # simple Primary HDU and append that first before
                    # we append the new Extension HDU.
                    phdu = PrimaryHDU()

                    super(HDUList, self).insert(0, phdu)
                    index = 1
        else:
            if isinstance(hdu, GroupsHDU):
                raise ValueError('A GroupsHDU must be inserted as a '
                                 'Primary HDU.')

            if isinstance(hdu, PrimaryHDU):
                # You passed a Primary HDU but we need an Extension HDU
                # so create an Extension HDU from the input Primary HDU.
                hdu = ImageHDU(hdu.data, hdu.header)

        super(HDUList, self).insert(index, hdu)
        hdu._new = True
        self._resize = True
        self._truncate = False
        # make sure the EXTEND keyword is in primary HDU if there is extension
        self.update_extend()

    def append(self, hdu):
        """
        Append a new HDU to the `HDUList`.

        Parameters
        ----------
        hdu : instance of _BaseHDU
            HDU to add to the `HDUList`.
        """

        if not isinstance(hdu, _BaseHDU):
            raise ValueError('HDUList can only append an HDU.')

        if len(self) > 0:
            if isinstance(hdu, GroupsHDU):
                raise ValueError(
                    "Can't append a GroupsHDU to a non-empty HDUList")

            if isinstance(hdu, PrimaryHDU):
                # You passed a Primary HDU but we need an Extension HDU
                # so create an Extension HDU from the input Primary HDU.
                # TODO: This isn't necessarily sufficient to copy the HDU;
                # _hdrLoc and friends need to be copied too.
                hdu = ImageHDU(hdu.data, hdu.header)
        else:
            if not isinstance(hdu, (PrimaryHDU, _NonstandardHDU)):
                # You passed in an Extension HDU but we need a Primary
                # HDU.
                # If you provided an ImageHDU then we can convert it to
                # a primary HDU and use that.
                if isinstance(hdu, ImageHDU):
                    hdu = PrimaryHDU(hdu.data, hdu.header)
                else:
                    # You didn't provide an ImageHDU so we create a
                    # simple Primary HDU and append that first before
                    # we append the new Extension HDU.
                    phdu = PrimaryHDU()
                    super(HDUList, self).append(phdu)

        super(HDUList, self).append(hdu)
        hdu._new = True
        self._resize = True
        self._truncate = False

        # make sure the EXTEND keyword is in primary HDU if there is extension
        self.update_extend()

    def index_of(self, key):
        """
        Get the index of an HDU from the `HDUList`.

        Parameters
        ----------
        key : int, str or tuple of (string, int)
           The key identifying the HDU.  If `key` is a tuple, it is of
           the form (`key`, `ver`) where `ver` is an ``EXTVER`` value
           that must match the HDU being searched for.

        Returns
        -------
        index : int
           The index of the HDU in the `HDUList`.
        """

        if _is_int(key):
            return key
        elif isinstance(key, tuple):
            _key, _ver = key
        else:
            _key = key
            _ver = None

        if not isinstance(_key, str):
            raise KeyError(key)
        _key = (_key.strip()).upper()

        nfound = 0
        found = None
        for idx, hdu in enumerate(self):
            name = hdu.name
            if isinstance(name, str):
                name = name.strip().upper()
            # 'PRIMARY' should always work as a reference to the first HDU
            if ((name == _key or (_key == 'PRIMARY' and idx == 0)) and
                (_ver is None or _ver == hdu._extver)):
                found = idx
                nfound += 1

        if (nfound == 0):
            raise KeyError('Extension %s not found.' % repr(key))
        elif (nfound > 1):
            raise KeyError('There are %d extensions of %s.'
                           % (nfound, repr(key)))
        else:
            return found

    def readall(self):
        """
        Read data of all HDUs into memory.
        """

        for hdu in self:
            if hdu.data is not None:
                continue

    @ignore_sigint
    def flush(self, output_verify='fix', verbose=False):
        """
        Force a write of the `HDUList` back to the file (for append and
        update modes only).

        Parameters
        ----------
        output_verify : str
            Output verification option.  Must be one of ``"fix"``,
            ``"silentfix"``, ``"ignore"``, ``"warn"``, or
            ``"exception"``.  See :ref:`verify` for more info.

        verbose : bool
            When `True`, print verbose messages
        """

        if self.__file.mode not in ('append', 'update', 'ostream'):
            warnings.warn("Flush for '%s' mode is not supported."
                          % self.__file.mode)
            return

        if self._save_backup and self.__file.mode in ('append', 'update'):
            filename = self.__file.name
            if os.path.exists(filename):
                # The the file doesn't actually exist anymore for some reason
                # then there's no point in trying to make a backup
                backup = filename + '.bak'
                idx = 1
                while os.path.exists(backup):
                    backup = filename + '.bak.' + str(idx)
                    idx += 1
                warnings.warn('Saving a backup of %s to %s.' %
                              (filename, backup))
                try:
                    shutil.copy(filename, backup)
                except IOError, e:
                    raise IOError('Failed to save backup to destination %s: '
                                  '%s' % (filename, str(e)))

        self.verify(option=output_verify)

        if self.__file.mode in ('append', 'ostream'):
            for hdu in self:
                if verbose:
                    try:
                        extver = str(hdu._header['extver'])
                    except KeyError:
                        extver = ''

                # only append HDU's which are "new"
                if hdu._new:
                    hdu._prewriteto(checksum=hdu._output_checksum)
                    try:
                        hdu._writeto(self.__file)
                        if verbose:
                            print 'append HDU', hdu.name, extver
                        hdu._new = False
                    finally:
                        hdu._postwriteto()
        elif self.__file.mode == 'update':
            self._flush_update()

    def update_extend(self):
        """
        Make sure that if the primary header needs the keyword
        ``EXTEND`` that it has it and it is correct.
        """

        if not len(self):
            return

        if not isinstance(self[0], PrimaryHDU):
            # A PrimaryHDU will be automatically inserted at some point, but it
            # might not have been added yet
            return

        hdr = self[0].header

        if 'EXTEND' in hdr:
            if len(self) > 1 and hdr['EXTEND'] == False:
                hdr['EXTEND'] = True
        elif len(self) > 1:
            if hdr['NAXIS'] == 0:
                hdr.set('EXTEND', True, after='NAXIS')
            else:
                n = hdr['NAXIS']
                hdr.set('EXTEND', True, after='NAXIS' + str(n))

    def writeto(self, fileobj, output_verify='exception', clobber=False,
                checksum=False):
        """
        Write the `HDUList` to a new file.

        Parameters
        ----------
        fileobj : file path, file object or file-like object
            File to write to.  If a file object, must be opened for
            append (ab+).

        output_verify : str
            Output verification option.  Must be one of ``"fix"``,
            ``"silentfix"``, ``"ignore"``, ``"warn"``, or
            ``"exception"``.  See :ref:`verify` for more info.

        clobber : bool
            When `True`, overwrite the output file if exists.

        checksum : bool
            When `True` adds both ``DATASUM`` and ``CHECKSUM`` cards
            to the headers of all HDU's written to the file.
        """

        if (len(self) == 0):
            warnings.warn("There is nothing to write.")
            return

        self.verify(option=output_verify)

        # check if the file object is closed
        closed = fileobj_closed(fileobj)
        fmode = fileobj_mode(fileobj) or 'ab+'
        filename = fileobj_name(fileobj)

        # check if the output file already exists
        if (isfile(fileobj) or
            isinstance(fileobj, (basestring, gzip.GzipFile))):
            if (os.path.exists(filename) and os.path.getsize(filename) != 0):
                if clobber:
                    warnings.warn("Overwriting existing file '%s'." % filename)
                    if not closed:
                        fileobj.close()
                    os.remove(filename)
                else:
                    raise IOError("File '%s' already exists." % filename)
        elif (hasattr(fileobj, 'len') and fileobj.len > 0):
            if clobber:
                warnings.warn("Overwriting existing file '%s'." % filename)
                name.truncate(0)
            else:
                raise IOError("File '%s' already exists." % filename)

        # make sure the EXTEND keyword is there if there is extension
        self.update_extend()

        mode = 'copyonwrite'
        for key, val in PYTHON_MODES.iteritems():
            if val == fmode:
                mode = key
                break

        hdulist = fitsopen(fileobj, mode=mode)

        for hdu in self:
            hdu._prewriteto(checksum=checksum)
            try:
                hdu._writeto(hdulist.__file)
            finally:
                hdu._postwriteto()
        hdulist.close(output_verify=output_verify, closed=closed)

    def close(self, output_verify='exception', verbose=False, closed=True):
        """
        Close the associated FITS file and memmap object, if any.

        Parameters
        ----------
        output_verify : str
            Output verification option.  Must be one of ``"fix"``,
            ``"silentfix"``, ``"ignore"``, ``"warn"``, or
            ``"exception"``.  See :ref:`verify` for more info.

        verbose : bool
            When `True`, print out verbose messages.

        closed : bool
            When `True`, close the underlying file object.
        """

        if self.__file:
            if self.__file.mode in ['append', 'update']:
                self.flush(output_verify=output_verify, verbose=verbose)

            if closed and hasattr(self.__file, 'close'):
                self.__file.close()

    def info(self, output=None):
        """
        Summarize the info of the HDUs in this `HDUList`.

        Note that this function prints its results to the console---it
        does not return a value.

        Parameters
        ----------
        output : file, optional
            A file-like object to write the output to.  If False, does not
            output to a file and instead returns a list of tuples representing
            the HDU info.  Writes to sys.stdout by default.
        """

        if output is None:
            output = sys.stdout

        if self.__file is None:
            name = '(No file associated with this HDUList)'
        else:
            name = self.__file.name

        results = ['Filename: %s' % name,
                   'No.    Name         Type      Cards   Dimensions   Format']

        format = '%-3d  %-10s  %-11s  %5d   %-10s   %s   %s'
        default = ('', '', 0, (), '', '')
        for idx, hdu in enumerate(self):
            summary = hdu._summary()
            if len(summary) < len(default):
                summary += default[len(summary):]
            summary = (idx,) + summary
            if output:
                results.append(format % summary)
            else:
                results.append(summary)

        if output:
            output.write('\n'.join(results))
            output.write('\n')
            output.flush()
        else:
            return results[2:]

    def filename(self):
        """
        Return the file name associated with the HDUList object if one exists.
        Otherwise returns None.

        Returns
        -------
        filename : a string containing the file name associated with the
                   HDUList object if an association exists.  Otherwise returns
                   None.
        """
        if self.__file is not None:
            if hasattr(self.__file, 'name'):
                return self.__file.name
        return None

    @classmethod
    def _readfrom(cls, fileobj=None, data=None, mode='readonly',
                  memmap=False, save_backup=False, **kwargs):
        """
        Provides the implementations from HDUList.fromfile and
        HDUList.fromstring, both of which wrap this method, as their
        implementations are largely the same.
        """

        if fileobj is not None:
            # instantiate a FITS file object (ffo)
            ffo = _File(fileobj, mode=mode, memmap=memmap)
            hdulist = cls(file=ffo)
        else:
            hdulist = cls()
            # This method is currently only called from HDUList.fromstring and
            # HDUList.fromfile.  If fileobj is None then this must be the
            # fromstring case; the data type of `data` will be checked in the
            # _BaseHDU.fromstring call.

        hdulist._save_backup = save_backup

        saved_compression_enabled = compressed.COMPRESSION_ENABLED

        try:
            if ('disable_image_compression' in kwargs and
                kwargs['disable_image_compression']):
                compressed.COMPRESSION_ENABLED = False

            if mode == 'ostream':
                # Output stream--not interested in reading/parsing the
                # HDUs--just writing to the output file
                return hdulist

            # read all HDUs
            while True:
                try:
                    if fileobj is not None:
                        try:
                            hdu = _BaseHDU.readfrom(ffo, **kwargs)
                        except EOFError:
                            break
                        except IOError, err:
                            if ffo.writeonly:
                                break
                            else:
                                raise
                    else:
                        if not data:
                            break
                        hdu = _BaseHDU.fromstring(data)
                        data = data[hdu._datLoc + hdu._datSpan:]
                    hdulist.append(hdu)
                    hdu._new = False
                    if 'checksum' in kwargs:
                        hdu._output_checksum = kwargs['checksum']
                # check in the case there is extra space after the last HDU or
                # corrupted HDU
                except (VerifyError, ValueError), err:
                    warnings.warn(
                        'Error validating header for HDU #%d (note: PyFITS '
                        'uses zero-based indexing).\n%s\n'
                        'There may be extra bytes after the last HDU or the '
                        'file is corrupted.' %
                        (len(hdulist), indent(str(err))), VerifyWarning)
                    break

            # If we're trying to read only and no header units were found,
            # raise and exception
            if mode in ('readonly', 'denywrite') and len(hdulist) == 0:
                raise IOError('Empty or corrupt FITS file')

            # initialize/reset attributes to be used in "update/append" mode
            hdulist._resize = False
            hdulist._truncate = False

        finally:
            compressed.COMPRESSION_ENABLED = saved_compression_enabled

        return hdulist

    def _verify(self, option='warn'):
        text = ''
        errs = _ErrList([], unit='HDU')

        # the first (0th) element must be a primary HDU
        if len(self) > 0 and (not isinstance(self[0], PrimaryHDU)) and \
                             (not isinstance(self[0], _NonstandardHDU)):
            err_text = "HDUList's 0th element is not a primary HDU."
            fix_text = 'Fixed by inserting one as 0th HDU.'

            def fix(self=self):
                self.insert(0, PrimaryHDU())

            err = self.run_option(option, err_text=err_text,
                                  fix_text=fix_text, fix=fix)
            errs.append(err)

        if len(self) > 1 and ('EXTEND' not in self[0].header or
                              self[0].header['EXTEND'] is not True):
            err_text = ('Primary HDU does not contain an EXTEND keyword '
                        'equal to T even though there are extension HDUs.')
            fix_text = 'Fixed by inserting or updating the EXTEND keyword.'

            def fix(header=self[0].header):
                naxis = header['NAXIS']
                if naxis == 0:
                    after = 'NAXIS'
                else:
                    after = 'NAXIS' + str(naxis)
                header.set('EXTEND', value=True, after=after)

            errs.append(self.run_option(option, err_text=err_text,
                                        fix_text=fix_text, fix=fix))

        # each element calls their own verify
        for idx, hdu in enumerate(self):
            if idx > 0 and (not isinstance(hdu, ExtensionHDU)):
                err_text = ("HDUList's element %s is not an extension HDU." %
                            str(idx))
                err = self.run_option(option, err_text=err_text, fixable=False)
                errs.append(err)

            else:
                result = hdu._verify(option)
                if result:
                    errs.append(result)
        return errs

    def _flush_update(self):
        """Implements flushing changes to a file in update mode."""

        for hdu in self:
            # Need to all _prewriteto() for each HDU first to determine if
            # resizing will be necessary
            hdu._prewriteto(checksum=hdu._output_checksum, inplace=True)

        try:
            self._wasresized()

            # if the HDUList is resized, need to write out the entire contents of
            # the hdulist to the file.
            if self._resize or self.__file.compression:
                self._flush_resize()
            else:
                # if not resized, update in place
                for hdu in self:
                    hdu._writeto(self.__file, inplace=True)

            # reset the modification attributes after updating
            for hdu in self:
                hdu._header._modified = False
        finally:
            for hdu in self:
                hdu._postwriteto()

    def _flush_resize(self):
        """
        Implements flushing changes in update mode when parts of one or more HDU
        need to be resized.
        """

        old_name = self.__file.name
        old_memmap = self.__file.memmap
        name = _tmp_name(old_name)

        if not self.__file.file_like:
            old_mode = os.stat(old_name).st_mode
            # The underlying file is an acutal file object.  The HDUList is
            # resized, so we need to write it to a tmp file, delete the
            # original file, and rename the tmp file to the original file.
            if self.__file.compression == 'gzip':
                new_file = gzip.GzipFile(name, mode='ab+')
            else:
                new_file = name

            hdulist = self.fromfile(new_file, mode='append')

            for hdu in self:
                hdu._writeto(hdulist.__file, inplace=True, copy=True)

            if sys.platform.startswith('win'):
                # Collect a list of open mmaps to the data; this well be used
                # later.  See below.
                mmaps = [(idx, _get_array_mmap(hdu.data), hdu.data)
                         for idx, hdu in enumerate(self) if hdu._data_loaded]

            hdulist.__file.close()
            self.__file.close()

            if sys.platform.startswith('win'):
                # Close all open mmaps to the data.  This is only necessary on
                # Windows, which will not allow a file to be renamed or deleted
                # until all handles to that file have been closed.
                for idx, mmap, arr in mmaps:
                    if mmap is not None:
                        mmap.close()

            os.remove(self.__file.name)

            # reopen the renamed new file with "update" mode
            os.rename(name, old_name)
            os.chmod(old_name, old_mode)

            if isinstance(new_file, gzip.GzipFile):
                old_file = gzip.GzipFile(old_name, mode='rb+')
            else:
                old_file = old_name

            ffo = _File(old_file, mode='update', memmap=old_memmap)

            self.__file = ffo

            for hdu in self:
                # Need to update the _file attribute and close any open mmaps
                # on each HDU
                if (hdu._data_loaded and
                    _get_array_mmap(hdu.data) is not None):
                    del hdu.data
                hdu._file = ffo

            if sys.platform.startswith('win'):
                # On Windows, all the original data mmaps were closed above.
                # However, it's possible that the user still has references to
                # the old data which would no longer work (possibly even cause
                # a segfault if they try to access it).  This replaces the
                # buffers used by the original arrays with the buffers of mmap
                # arrays created from the new file.  This seems to work, but
                # it's a flaming hack and carries no guarantees that it won't
                # lead to odd behavior in practice.  Better to just not keep
                # references to data from files that had to be resized upon
                # flushing (on Windows--again, this is no problem on Linux).
                for idx, mmap, arr in mmaps:
                    if mmap is not None:
                        arr.data = self[idx].data.data
                del mmaps  # Just to be sure

        else:
            # The underlying file is not a file object, it is a file like
            # object.  We can't write out to a file, we must update the file
            # like object in place.  To do this, we write out to a temporary
            # file, then delete the contents in our file like object, then
            # write the contents of the temporary file to the now empty file
            # like object.
            self.writeto(name)
            hdulist = self.fromfile(name)
            ffo = self.__file

            ffo.truncate(0)
            ffo.seek(0)

            for hdu in hdulist:
                hdu._writeto(ffo, inplace=True, copy=True)

            # Close the temporary file and delete it.
            hdulist.close()
            os.remove(hdulist.__file.name)

        # reset the resize attributes after updating
        self._resize = False
        self._truncate = False
        for hdu in self:
            hdu._header._modified = False
            hdu._new = False
            hdu._file = ffo

    def _wasresized(self, verbose=False):
        """
        Determine if any changes to the HDUList will require a file resize
        when flushing the file.

        Side effect of setting the objects _resize attribute.
        """

        if not self._resize:

            # determine if any of the HDU is resized
            for hdu in self:
                # Header:
                nbytes = len(str(hdu._header))
                if nbytes != (hdu._datLoc - hdu._hdrLoc):
                    self._resize = True
                    self._truncate = False
                    if verbose:
                        print 'One or more header is resized.'
                    break

                # Data:
                if not hdu._data_loaded or hdu.data is None:
                    continue

                nbytes = hdu.size
                nbytes = nbytes + _pad_length(nbytes)
                if nbytes != hdu._datSpan:
                    self._resize = True
                    self._truncate = False
                    if verbose:
                        print 'One or more data area is resized.'
                    break

            if self._truncate:
                try:
                    self.__file.truncate(hdu._datLoc + hdu._datSpan)
                except IOError:
                    self._resize = True
                self._truncate = False

        return self._resize



from __future__ import division, with_statement

import platform

import numpy as np

import pyfits
from pyfits.tests import PyfitsTestCase

from nose.tools import assert_equal, assert_true

class TestUintFunctions(PyfitsTestCase):
    def test_uint16(self):
        hdu = pyfits.PrimaryHDU(np.array([-3,-2,-1,0,1,2,3]))
        hdu.scale('int16', '', bzero=2**15)
        hdu.writeto(self.temp('tempfile.fits'))
        hdul = pyfits.open(self.temp('tempfile.fits'), uint=True)
        assert_equal(hdul[0].data.dtype, np.uint16)
        assert_equal(np.all(hdul[0].data ==
                         np.array([(2**16)-3, (2**16)-2, (2**16)-1, 0, 1, 2, 3],
                                  dtype=np.uint16)), True)
        hdul.writeto(self.temp('tempfile1.fits'))
        hdul1 = pyfits.open(self.temp('tempfile1.fits'), uint16=True)
        assert_equal(np.all(hdul[0].data == hdul1[0].data), True)
        assert_equal(hdul[0].section[:1].dtype.name, 'uint16')
        assert_true((hdul[0].section[:1] == hdul[0].data[:1]).all())
        hdul.close()
        hdul1.close()

    def test_uint32(self):
        hdu = pyfits.PrimaryHDU(np.array([-3, -2, -1, 0, 1, 2, 3]))
        hdu.scale('int32', '', bzero=2**31)
        hdu.writeto(self.temp('tempfile.fits'))
        hdul = pyfits.open(self.temp('tempfile.fits'), uint=True)
        assert_equal(hdul[0].data.dtype, np.uint32)
        assert_equal(np.all(hdul[0].data ==
                         np.array([(2**32)-3, (2**32)-2, (2**32)-1, 0, 1, 2, 3],
                         dtype=np.uint32)), True)
        hdul.writeto(self.temp('tempfile1.fits'))
        hdul1 = pyfits.open(self.temp('tempfile1.fits'), uint=True)
        assert_equal(np.all(hdul[0].data == hdul1[0].data), True)
        assert_equal(hdul[0].section[:1].dtype.name, 'uint32')
        assert_true((hdul[0].section[:1] == hdul[0].data[:1]).all())
        hdul.close()
        hdul1.close()

    def test_uint64(self):
        if platform.architecture()[0] == '64bit':
            hdu = pyfits.PrimaryHDU(np.array([-3,-2,-1,0,1,2,3]))
            hdu.scale('int64', '', bzero=2**63)
            hdu.writeto(self.temp('tempfile.fits'))
            hdul = pyfits.open(self.temp('tempfile.fits'), uint=True)
            assert_equal(hdul[0].data.dtype, np.uint64)
            assert_equal(np.all(hdul[0].data ==
                             np.array([(2**64)-3,(2**64)-2,(2**64)-1,0,1,2,3],
                             dtype=np.uint64)), True)
            hdul.writeto(self.temp('tempfile1.fits'))
            hdul1 = pyfits.open(self.temp('tempfile1.fits'), uint=True)
            assert_equal(np.all(hdul[0].data == hdul1[0].data), True)
            assert_equal(hdul[0].section[:1].dtype.name, 'uint64')
            assert_true((hdul[0].section[:1] == hdul[0].data[:1]).all())
            hdul.close()
            hdul1.close()

    def test_uint_compressed(self):
        hdu = pyfits.CompImageHDU(np.array([-3, -2, -1, 0, 1, 2, 3]))
        hdu.scale('int32', '', bzero=2**31)
        hdu.writeto(self.temp('temp.fits'))
        with pyfits.open(self.temp('temp.fits'), uint=True) as hdul:
            assert_equal(hdul[1].data.dtype, np.uint32)
            assert_true(
                (hdul[1].data ==
                 np.array([(2**32)-3, (2**32)-2, (2**32)-1, 0, 1, 2, 3],
                          dtype=np.uint32)).all())
            hdul.writeto(self.temp('temp2.fits'))
            with pyfits.open(self.temp('temp2.fits'), uint=True) as hdul2:
                assert_true((hdul[1].data == hdul2[1].data).all())
                # TODO: Enable these lines if CompImageHDUs ever grow .section
                # support
                #assert_equal(hdul[1].section[:1].dtype.name, 'uint32')
                #assert_true((hdul[1].section[:1] == hdul[1].data[:1]).all())



import textwrap

import numpy as np

import pyfits
from pyfits.column import Column
from pyfits.diff import *
from pyfits.hdu import HDUList, PrimaryHDU, ImageHDU
from pyfits.hdu.table import new_table
from pyfits.header import Header
from pyfits.tests import PyfitsTestCase

from nose.tools import (assert_true, assert_false, assert_equal,
                        assert_not_equal)


class TestDiff(PyfitsTestCase):
    def test_identical_headers(self):
        ha = Header([('A', 1), ('B', 2), ('C', 3)])
        hb = ha.copy()
        assert_true(HeaderDiff(ha, hb).identical)

    def test_slightly_different_headers(self):
        ha = Header([('A', 1), ('B', 2), ('C', 3)])
        hb = ha.copy()
        hb['C'] = 4
        assert_false(HeaderDiff(ha, hb).identical)

    def test_common_keywords(self):
        ha = Header([('A', 1), ('B', 2), ('C', 3)])
        hb = ha.copy()
        hb['C'] = 4
        hb['D'] = (5, 'Comment')
        assert_equal(HeaderDiff(ha, hb).common_keywords, ['A', 'B', 'C'])

    def test_different_keyword_count(self):
        ha = Header([('A', 1), ('B', 2), ('C', 3)])
        hb = ha.copy()
        del hb['B']
        diff = HeaderDiff(ha, hb)
        assert_false(diff.identical)
        assert_equal(diff.diff_keyword_count, (3, 2))

        # But make sure the common keywords are at least correct
        assert_equal(diff.common_keywords, ['A', 'C'])

    def test_different_keywords(self):
        ha = Header([('A', 1), ('B', 2), ('C', 3)])
        hb = ha.copy()
        hb['C'] = 4
        hb['D'] = (5, 'Comment')
        ha['E'] = (6, 'Comment')
        ha['F'] = (7, 'Comment')
        diff = HeaderDiff(ha, hb)
        assert_false(diff.identical)
        assert_equal(diff.diff_keywords, (['E', 'F'], ['D']))

    def test_different_keyword_values(self):
        ha = Header([('A', 1), ('B', 2), ('C', 3)])
        hb = ha.copy()
        hb['C'] = 4
        diff = HeaderDiff(ha, hb)
        assert_false(diff.identical)
        assert_equal(diff.diff_keyword_values, {'C': [(3, 4)]})

    def test_different_keyword_comments(self):
        ha = Header([('A', 1), ('B', 2), ('C', 3, 'comment 1')])
        hb = ha.copy()
        hb.comments['C'] = 'comment 2'
        diff = HeaderDiff(ha, hb)
        assert_false(diff.identical)
        assert_equal(diff.diff_keyword_comments,
                     {'C': [('comment 1', 'comment 2')]})

    def test_different_keyword_values_with_duplicate(self):
        ha = Header([('A', 1), ('B', 2), ('C', 3)])
        hb = ha.copy()
        ha.append(('C', 4))
        hb.append(('C', 5))
        diff = HeaderDiff(ha, hb)
        assert_false(diff.identical)
        assert_equal(diff.diff_keyword_values, {'C': [None, (4, 5)]})

    def test_asymmetric_duplicate_keywords(self):
        ha = Header([('A', 1), ('B', 2), ('C', 3)])
        hb = ha.copy()
        ha.append(('A', 2, 'comment 1'))
        ha.append(('A', 3, 'comment 2'))
        hb.append(('B', 4, 'comment 3'))
        hb.append(('C', 5, 'comment 4'))
        diff = HeaderDiff(ha, hb)
        assert_false(diff.identical)
        assert_equal(diff.diff_keyword_values, {})
        assert_equal(diff.diff_duplicate_keywords,
                     {'A': (3, 1), 'B': (1, 2), 'C': (1, 2)})

    def test_floating_point_tolerance(self):
        ha = Header([('A', 1), ('B', 2.00001), ('C', 3.000001)])
        hb = ha.copy()
        hb['B'] = 2.00002
        hb['C'] = 3.000002
        diff = HeaderDiff(ha, hb)
        assert_false(diff.identical)
        assert_equal(diff.diff_keyword_values,
                     {'B': [(2.00001, 2.00002)], 'C': [(3.000001, 3.000002)]})
        diff = HeaderDiff(ha, hb, tolerance=1e-6)
        assert_false(diff.identical)
        assert_equal(diff.diff_keyword_values, {'B': [(2.00001, 2.00002)]})

    def test_ignore_blanks(self):
        pyfits.STRIP_HEADER_WHITESPACE = False
        try:
            ha = Header([('A', 1), ('B', 2), ('C', 'A       ')])
            hb = ha.copy()
            hb['C'] = 'A'
            assert_not_equal(ha['C'], hb['C'])

            diff = HeaderDiff(ha, hb)
            # Trailing blanks are ignored by default
            assert_true(diff.identical)
            assert_equal(diff.diff_keyword_values, {})

            # Don't ignore blanks
            diff = HeaderDiff(ha, hb, ignore_blanks=False)
            assert_false(diff.identical)
            assert_equal(diff.diff_keyword_values, {'C': [('A       ', 'A')]})
        finally:
            pyfits.STRIP_HEADER_WHITESPACE = True

    def test_ignore_blank_cards(self):
        """Test for #152--ignore blank cards."""

        ha = Header([('A', 1), ('B', 2), ('C', 3)])
        hb = Header([('A', 1), ('', ''), ('B', 2), ('', ''), ('C', 3)])
        hc = ha.copy()
        hc.append()
        hc.append()

        # We now have a header with interleaved blanks, and a header with end
        # blanks, both of which should ignore the blanks
        assert_true(HeaderDiff(ha, hb).identical)
        assert_true(HeaderDiff(ha, hc).identical)
        assert_true(HeaderDiff(hb, hc).identical)

        assert_false(HeaderDiff(ha, hb, ignore_blank_cards=False).identical)
        assert_false(HeaderDiff(ha, hc, ignore_blank_cards=False).identical)

        # Both hb and hc have the same number of blank cards; since order is
        # currently ignored, these should still be identical even if blank
        # cards are not ignored
        assert_true(HeaderDiff(hb, hc, ignore_blank_cards=False).identical)

        hc.append()
        # But now there are different numbers of blanks, so they should not be
        # ignored:
        assert_false(HeaderDiff(hb, hc, ignore_blank_cards=False).identical)

    def test_ignore_keyword_values(self):
        ha = Header([('A', 1), ('B', 2), ('C', 3)])
        hb = ha.copy()
        hb['B'] = 4
        hb['C'] = 5
        diff = HeaderDiff(ha, hb, ignore_keywords=['*'])
        assert_true(diff.identical)
        diff = HeaderDiff(ha, hb, ignore_keywords=['B'])
        assert_false(diff.identical)
        assert_equal(diff.diff_keyword_values, {'C': [(3, 5)]})

        report = diff.report()
        assert_true('Keyword B        has different values' not in report)
        assert_true('Keyword C        has different values' in report)

        # Test case-insensitivity
        diff = HeaderDiff(ha, hb, ignore_keywords=['b'])
        assert_false(diff.identical)
        assert_equal(diff.diff_keyword_values, {'C': [(3, 5)]})

    def test_ignore_keyword_comments(self):
        ha = Header([('A', 1, 'A'), ('B', 2, 'B'), ('C', 3, 'C')])
        hb = ha.copy()
        hb.comments['B'] = 'D'
        hb.comments['C'] = 'E'
        diff = HeaderDiff(ha, hb, ignore_comments=['*'])
        assert_true(diff.identical)
        diff = HeaderDiff(ha, hb, ignore_comments=['B'])
        assert_false(diff.identical)
        assert_equal(diff.diff_keyword_comments, {'C': [('C', 'E')]})

        report = diff.report()
        assert_true('Keyword B        has different comments' not in report)
        assert_true('Keyword C        has different comments' in report)

        # Test case-insensitivity
        diff = HeaderDiff(ha, hb, ignore_comments=['b'])
        assert_false(diff.identical)
        assert_equal(diff.diff_keyword_comments, {'C': [('C', 'E')]})

    def test_trivial_identical_images(self):
        ia = np.arange(100).reshape((10, 10))
        ib = np.arange(100).reshape((10, 10))
        diff = ImageDataDiff(ia, ib)
        assert_true(diff.identical)
        assert_equal(diff.diff_total, 0)

    def test_identical_within_tolerance(self):
        ia = np.ones((10, 10)) - 0.00001
        ib = np.ones((10, 10)) - 0.00002
        diff = ImageDataDiff(ia, ib, tolerance=1.0e-4)
        assert_true(diff.identical)
        assert_equal(diff.diff_total, 0)

    def test_identical_comp_image_hdus(self):
        """Regression test for #189.

        For this test we mostly just care that comparing to compressed images
        does not crash, and returns the correct results.  Two compressed images
        will be considered identical if the decompressed data is the same.
        Obviously we test whether or not the same compression was used by
        looking for (or ignoring) header differences.
        """

        data = np.arange(100.0).reshape((10, 10))
        hdu = pyfits.CompImageHDU(data=data)
        hdu.writeto(self.temp('test.fits'))
        hdula = pyfits.open(self.temp('test.fits'))
        hdulb = pyfits.open(self.temp('test.fits'))
        diff = FITSDiff(hdula, hdulb)
        assert_true(diff.identical)

    def test_different_dimensions(self):
        ia = np.arange(100).reshape((10, 10))
        ib = np.arange(100) - 1

        # Although ib could be reshaped into the same dimensions, for now the
        # data is not compared anyways
        diff = ImageDataDiff(ia, ib)
        assert_false(diff.identical)
        assert_equal(diff.diff_dimensions, ((10, 10), (100,)))
        assert_equal(diff.diff_total, 0)

        report = diff.report()
        assert_true('Data dimensions differ' in report)
        assert_true('a: 10 x 10' in report)
        assert_true('b: 100' in report)
        assert_true('No further data comparison performed.')

    def test_different_pixels(self):
        ia = np.arange(100).reshape((10, 10))
        ib = np.arange(100).reshape((10, 10))
        ib[0,0] = 10
        ib[5,5] = 20
        diff = ImageDataDiff(ia, ib)
        assert_false(diff.identical)
        assert_equal(diff.diff_dimensions, ())
        assert_equal(diff.diff_total, 2)
        assert_equal(diff.diff_ratio, 0.02)
        assert_equal(diff.diff_pixels, [((0, 0), (0, 10)), ((5, 5), (55, 20))])

    def test_identical_tables(self):
        c1 = Column('A', format='L', array=[True, False])
        c2 = Column('B', format='X', array=[[0], [1]])
        c3 = Column('C', format='4I', dim='(2, 2)',
                    array=[[0, 1, 2, 3], [4, 5, 6, 7]])
        c4 = Column('D', format='J', bscale=2.0, array=[0, 1])
        c5 = Column('E', format='A3', array=['abc', 'def'])
        c6 = Column('F', format='E', unit='m', array=[0.0, 1.0])
        c7 = Column('G', format='D', bzero=-0.1, array=[0.0, 1.0])
        c8 = Column('H', format='C', array=[0.0+1.0j, 2.0+3.0j])
        c9 = Column('I', format='M', array=[4.0+5.0j, 6.0+7.0j])
        c10 = Column('J', format='PI(2)', array=[[0, 1], [2, 3]])

        columns = [c1, c2, c3, c4, c5, c6, c7, c8, c9, c10]

        ta = new_table(columns)
        tb = new_table([c.copy() for c in columns])

        diff = TableDataDiff(ta.data, tb.data)
        assert_true(diff.identical)
        assert_equal(len(diff.common_columns), 10)
        assert_equal(diff.common_column_names,
                     set(['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']))
        assert_equal(diff.diff_ratio, 0)
        assert_equal(diff.diff_total, 0)

    def test_diff_empty_tables(self):
        """
        Regression test for #178.

        Ensure that diffing tables containing empty data doesn't crash.
        """

        c1 = Column('D', format='J')
        c2 = Column('E', format='J')
        thdu = new_table([c1, c2], nrows=0)

        hdula = pyfits.HDUList([thdu])
        hdulb = pyfits.HDUList([thdu])

        diff = FITSDiff(hdula, hdulb)
        assert_true(diff.identical)

    def test_ignore_table_fields(self):
        c1 = Column('A', format='L', array=[True, False])
        c2 = Column('B', format='X', array=[[0], [1]])
        c3 = Column('C', format='4I', dim='(2, 2)',
                    array=[[0, 1, 2, 3], [4, 5, 6, 7]])

        c4 = Column('B', format='X', array=[[1], [0]])
        c5 = Column('C', format='4I', dim='(2, 2)',
                    array=[[1, 2, 3, 4], [5, 6, 7, 8]])

        ta = new_table([c1, c2, c3])
        tb = new_table([c1, c4, c5])

        diff = TableDataDiff(ta.data, tb.data, ignore_fields=['B', 'C'])
        assert_true(diff.identical)

        # The only common column should be c1
        assert_equal(len(diff.common_columns), 1)
        assert_equal(diff.common_column_names, set(['a']))
        assert_equal(diff.diff_ratio, 0)
        assert_equal(diff.diff_total, 0)

    def test_different_table_field_names(self):
        ca = Column('A', format='L', array=[True, False])
        cb = Column('B', format='L', array=[True, False])
        cc = Column('C', format='L', array=[True, False])

        ta = new_table([ca, cb])
        tb = new_table([ca, cc])

        diff = TableDataDiff(ta.data, tb.data)

        assert_false(diff.identical)
        assert_equal(len(diff.common_columns), 1)
        assert_equal(diff.common_column_names, set(['a']))
        assert_equal(diff.diff_column_names, (['B'], ['C']))
        assert_equal(diff.diff_ratio, 0)
        assert_equal(diff.diff_total, 0)

    def test_different_table_field_counts(self):
        """
        Test tables with some common columns, but different number of columns
        overall.
        """

        ca = Column('A', format='L', array=[True, False])
        cb = Column('B', format='L', array=[True, False])
        cc = Column('C', format='L', array=[True, False])

        ta = new_table([cb])
        tb = new_table([ca, cb, cc])

        diff = TableDataDiff(ta.data, tb.data)

        assert_false(diff.identical)
        assert_equal(diff.diff_column_count, (1, 3))
        assert_equal(len(diff.common_columns), 1)
        assert_equal(diff.common_column_names, set(['b']))
        assert_equal(diff.diff_column_names, ([], ['A', 'C']))
        assert_equal(diff.diff_ratio, 0)
        assert_equal(diff.diff_total, 0)

    def test_different_table_rows(self):
        """
        Test tables taht are otherwise identical but one has more rows than the
        other.
        """

        ca1 = Column('A', format='L', array=[True, False])
        cb1 = Column('B', format='L', array=[True, False])
        ca2 = Column('A', format='L', array=[True, False, True])
        cb2 = Column('B', format='L', array=[True, False, True])

        ta = new_table([ca1, cb1])
        tb = new_table([ca2, cb2])

        diff = TableDataDiff(ta.data, tb.data)

        assert_false(diff.identical)
        assert_equal(diff.diff_column_count, ())
        assert_equal(len(diff.common_columns), 2)
        assert_equal(diff.diff_rows, (2, 3))
        assert_equal(diff.diff_values, [])

        report = diff.report()

        assert_true('Table rows differ' in report)
        assert_true('a: 2' in report)
        assert_true('b: 3' in report)
        assert_true('No further data comparison performed.')

    def test_different_table_data(self):
        """
        Test diffing table data on columns of several different data formats
        and dimensions.
        """

        ca1 = Column('A', format='L', array=[True, False])
        ca2 = Column('B', format='X', array=[[0], [1]])
        ca3 = Column('C', format='4I', dim='(2, 2)',
                     array=[[0, 1, 2, 3], [4, 5, 6, 7]])
        ca4 = Column('D', format='J', bscale=2.0, array=[0.0, 2.0])
        ca5 = Column('E', format='A3', array=['abc', 'def'])
        ca6 = Column('F', format='E', unit='m', array=[0.0, 1.0])
        ca7 = Column('G', format='D', bzero=-0.1, array=[0.0, 1.0])
        ca8 = Column('H', format='C', array=[0.0+1.0j, 2.0+3.0j])
        ca9 = Column('I', format='M', array=[4.0+5.0j, 6.0+7.0j])
        ca10 = Column('J', format='PI(2)', array=[[0, 1], [2, 3]])

        cb1 = Column('A', format='L', array=[False, False])
        cb2 = Column('B', format='X', array=[[0], [0]])
        cb3 = Column('C', format='4I', dim='(2, 2)',
                     array=[[0, 1, 2, 3], [5, 6, 7, 8]])
        cb4 = Column('D', format='J', bscale=2.0, array=[2.0, 2.0])
        cb5 = Column('E', format='A3', array=['abc', 'ghi'])
        cb6 = Column('F', format='E', unit='m', array=[1.0, 2.0])
        cb7 = Column('G', format='D', bzero=-0.1, array=[2.0, 3.0])
        cb8 = Column('H', format='C', array=[1.0+1.0j, 2.0+3.0j])
        cb9 = Column('I', format='M', array=[5.0+5.0j, 6.0+7.0j])
        cb10 = Column('J', format='PI(2)', array=[[1, 2], [3, 4]])

        ta = new_table([ca1, ca2, ca3, ca4, ca5, ca6, ca7, ca8, ca9, ca10])
        tb = new_table([cb1, cb2, cb3, cb4, cb5, cb6, cb7, cb8, cb9, cb10])

        diff = TableDataDiff(ta.data, tb.data, numdiffs=20)
        assert_false(diff.identical)
        # The column definitions are the same, but not the column values
        assert_equal(diff.diff_columns, ())
        assert_equal(diff.diff_values[0], (('A', 0), (True, False)))
        assert_equal(diff.diff_values[1], (('B', 1), ([1], [0])))
        assert_equal(diff.diff_values[2][0], ('C', 1))
        assert_true((diff.diff_values[2][1][0] == [[4, 5], [6, 7]]).all())
        assert_true((diff.diff_values[2][1][1] == [[5, 6], [7, 8]]).all())
        assert_equal(diff.diff_values[3], (('D', 0), (0, 2.0)))
        assert_equal(diff.diff_values[4], (('E', 1), ('def', 'ghi')))
        assert_equal(diff.diff_values[5], (('F', 0), (0.0, 1.0)))
        assert_equal(diff.diff_values[6], (('F', 1), (1.0, 2.0)))
        assert_equal(diff.diff_values[7], (('G', 0), (0.0, 2.0)))
        assert_equal(diff.diff_values[8], (('G', 1), (1.0, 3.0)))
        assert_equal(diff.diff_values[9], (('H', 0), (0.0+1.0j, 1.0+1.0j)))
        assert_equal(diff.diff_values[10], (('I', 0), (4.0+5.0j, 5.0+5.0j)))
        assert_equal(diff.diff_values[11][0], ('J', 0))
        assert_true((diff.diff_values[11][1][0] == [0, 1]).all())
        assert_true((diff.diff_values[11][1][1] == [1, 2]).all())
        assert_equal(diff.diff_values[12][0], ('J', 1))
        assert_true((diff.diff_values[12][1][0] == [2, 3]).all())
        assert_true((diff.diff_values[12][1][1] == [3, 4]).all())

        assert_equal(diff.diff_total, 13)
        assert_equal(diff.diff_ratio, 0.65)

    def test_identical_files_basic(self):
        """Test identicality of two simple, extensionless files."""

        a = np.arange(100).reshape((10, 10))
        hdu = PrimaryHDU(data=a)
        hdu.writeto(self.temp('testa.fits'))
        hdu.writeto(self.temp('testb.fits'))
        diff = FITSDiff(self.temp('testa.fits'), self.temp('testb.fits'))
        assert_true(diff.identical)

        report = diff.report()
        # Primary HDUs should contain no differences
        assert_true('Primary HDU' not in report)
        assert_true('Extension HDU' not in report)
        assert_true('No differences found.' in report)

    def test_partially_identical_files1(self):
        """
        Test files that have some identical HDUs but a different extension
        count.
        """

        a = np.arange(100).reshape((10, 10))
        phdu = PrimaryHDU(data=a)
        ehdu = ImageHDU(data=a)
        hdula = HDUList([phdu, ehdu])
        hdulb = HDUList([phdu, ehdu, ehdu])
        diff = FITSDiff(hdula, hdulb)
        assert_false(diff.identical)
        assert_equal(diff.diff_hdu_count, (2, 3))

        # diff_hdus should be empty, since the third extension in hdulb
        # has nothing to compare against
        assert_equal(diff.diff_hdus, [])

        report = diff.report()
        assert_true('Files contain different numbers of HDUs' in report)
        assert_true('a: 2\n b: 3' in report)
        assert_true('No differences found between common HDUs' in report)

    def test_partially_identical_files2(self):
        """
        Test files that have some identical HDUs but one different HDU.
        """

        a = np.arange(100).reshape((10, 10))
        phdu = PrimaryHDU(data=a)
        ehdu = ImageHDU(data=a)
        ehdu2 = ImageHDU(data=(a + 1))
        hdula = HDUList([phdu, ehdu, ehdu])
        hdulb = HDUList([phdu, ehdu2, ehdu])
        diff = FITSDiff(hdula, hdulb)

        assert_false(diff.identical)
        assert_equal(diff.diff_hdu_count, ())
        assert_equal(len(diff.diff_hdus), 1)
        assert_equal(diff.diff_hdus[0][0], 1)

        hdudiff = diff.diff_hdus[0][1]
        assert_false(hdudiff.identical)
        assert_equal(hdudiff.diff_extnames, ())
        assert_equal(hdudiff.diff_extvers, ())
        assert_equal(hdudiff.diff_extension_types, ())
        assert_true(hdudiff.diff_headers.identical)
        assert_false(hdudiff.diff_data is None)

        datadiff = hdudiff.diff_data
        assert_true(isinstance(datadiff, ImageDataDiff))
        assert_false(datadiff.identical)
        assert_equal(datadiff.diff_dimensions, ())
        assert_equal(datadiff.diff_pixels,
                     [((0, y), (y, y + 1)) for y in range(10)])
        assert_equal(datadiff.diff_ratio, 1.0)
        assert_equal(datadiff.diff_total, 100)

        report = diff.report()
        # Primary HDU and 2nd extension HDU should have no differences
        assert_true('Primary HDU' not in report)
        assert_true('Extension HDU 2' not in report)
        assert_true('Extension HDU 1' in report)

        assert_true('Headers contain differences' not in report)
        assert_true('Data contains differences' in report)
        for y in range(10):
            assert_true('Data differs at [%d, 1]' % (y + 1) in report)
        assert_true('100 different pixels found (100.00% different).' in
                    report)

    def test_diff_nans(self):
        """Regression test for #204."""

        # First test some arrays that should be equivalent....
        arr = np.empty((10, 10), dtype=np.float64)
        arr[:5] = 1.0
        arr[5:] = np.nan
        arr2 = arr.copy()

        table = np.rec.array([(1.0, 2.0), (3.0, np.nan), (np.nan, np.nan)],
                             names=['cola', 'colb']).view(pyfits.FITS_rec)
        table2 = table.copy()

        assert_true(ImageDataDiff(arr, arr2).identical)
        assert_true(TableDataDiff(table, table2).identical)

        # Now let's introduce some differences, where there are nans and where
        # there are not nans
        arr2[0][0] = 2.0
        arr2[5][0] = 2.0
        table2[0][0] = 2.0
        table2[1][1] = 2.0

        diff = ImageDataDiff(arr, arr2)
        assert_false(diff.identical)
        assert_equal(diff.diff_pixels[0], ((0, 0), (1.0, 2.0)))
        assert_equal(diff.diff_pixels[1][0], (5, 0))
        assert_true(np.isnan(diff.diff_pixels[1][1][0]))
        assert_equal(diff.diff_pixels[1][1][1], 2.0)

        diff = TableDataDiff(table, table2)
        assert_false(diff.identical)
        assert_equal(diff.diff_values[0], (('cola', 0), (1.0, 2.0)))
        assert_equal(diff.diff_values[1][0], ('colb', 1))
        assert_true(np.isnan(diff.diff_values[1][1][0]))
        assert_equal(diff.diff_values[1][1][1], 2.0)

        # What about in Headers?

        h = pyfits.Header([('A', 1), ('B', np.nan)])
        h2 = h.copy()

        assert_true(HeaderDiff(h, h2).identical)

        h2['B'] = 1.0

        diff = HeaderDiff(h, h2)
        assert_false(diff.identical)
        assert_true(np.isnan(diff.diff_keyword_values['B'][0][0]))
        assert_equal(diff.diff_keyword_values['B'][0][1], 1.0)



"""Test utility functions."""

import sys
import warnings

from pyfits.util import StringIO


class CaptureStdio(object):
    """
    A simple context manager for redirecting stdout and stderr to a StringIO
    buffer.
    """

    def __init__(self, stdout=True, stderr=True):
        self.stdout = StringIO()
        self.stderr = StringIO()

    def __enter__(self):
        self._original_stdout = sys.stdout
        self._original_stderr = sys.stderr
        sys.stdout = self.stdout
        sys.stderr = self.stderr
        return self.stdout, self.stderr

    def __exit__(self, *args, **kwargs):
        sys.stdout = self._original_stdout
        sys.stderr = self._original_stderr
        self.stdout.close()
        self.stderr.close()


if hasattr(warnings, 'catch_warnings'):
    catch_warnings = warnings.catch_warnings
else:
    # For Python2.5, backport the catch_warnings context manager
    class WarningMessage(object):

        """Holds the result of a single showwarning() call."""

        _WARNING_DETAILS = ("message", "category", "filename", "lineno",
                            "file", "line")

        def __init__(self, message, category, filename, lineno, file=None,
                        line=None):
            local_values = locals()
            for attr in self._WARNING_DETAILS:
                setattr(self, attr, local_values[attr])
            self._category_name = category.__name__ if category else None

        def __str__(self):
            return ("{message : %r, category : %r, filename : %r, lineno : %s,"
                    " line : %r}" % (self.message, self._category_name,
                                     self.filename, self.lineno, self.line))
    class catch_warnings(object):

        """A context manager that copies and restores the warnings filter upon
        exiting the context.

        The 'record' argument specifies whether warnings should be captured by
        a custom implementation of warnings.showwarning() and be appended to a
        list returned by the context manager. Otherwise None is returned by the
        context manager. The objects appended to the list are arguments whose
        attributes mirror the arguments to showwarning().

        The 'module' argument is to specify an alternative module to the module
        named 'warnings' and imported under that name. This argument is only
        useful when testing the warnings module itself.

        """

        def __init__(self, record=False, module=None):
            """Specify whether to record warnings and if an alternative module
            should be used other than sys.modules['warnings'].

            For compatibility with Python 3.0, please consider all arguments to
            be keyword-only.

            """
            self._record = record
            self._module = (sys.modules['warnings']
                            if module is None else module)
            self._entered = False

        def __repr__(self):
            args = []
            if self._record:
                args.append("record=True")
            if self._module is not sys.modules['warnings']:
                args.append("module=%r" % self._module)
            name = type(self).__name__
            return "%s(%s)" % (name, ", ".join(args))

        def __enter__(self):
            if self._entered:
                raise RuntimeError("Cannot enter %r twice" % self)
            self._entered = True
            self._filters = self._module.filters
            self._module.filters = self._filters[:]
            self._showwarning = self._module.showwarning
            if self._record:
                log = []
                def showwarning(*args, **kwargs):
                    log.append(WarningMessage(*args, **kwargs))
                self._module.showwarning = showwarning
                return log
            else:
                return None

        def __exit__(self, *exc_info):
            if not self._entered:
                raise RuntimeError("Cannot exit %r without entering first" %
                                   self)
            self._module.filters = self._filters
            self._module.showwarning = self._showwarning


class ignore_warnings(catch_warnings):
    def __enter__(self):
        retval = super(ignore_warnings, self).__enter__()
        warnings.simplefilter('ignore')
        return retval



"""
fitscheck is a command line script based on pyfits for verifying
and updating the CHECKSUM and DATASUM keywords of .fits files.
Fitscheck can also detect and often fix other FITS standards
violations.  fitscheck facilitates re-writing the non-standard
checksums originally generated by pyfits with standard checksums which
will interoperate with cfitsio.

fitscheck will refuse to write new checksums if the checksum keywords
are missing or their values are bad.  Use --force to write new
checksums regardless of whether or not they currently exist or pass.
Use --ignore-missing to tolerate missing checksum keywords without
comment.

Example uses of fitscheck:

1. Verify and update checksums, tolerating non-standard checksums, updating to
   standard checksum::

    $ fitscheck --checksum either --write *.fits

2. Write new checksums,  even if existing checksums are bad or missing::

    $ fitscheck --write --force *.fits

3. Verify standard checksums and FITS compliance without changing the files::

    $ fitscheck --compliance *.fits

4. Verify original nonstandard checksums only::

    $ fitscheck --checksum nonstandard *.fits

5. Only check and fix compliance problems,  ignoring checksums::

    $ fitscheck --checksum none --compliance --write *.fits

6. Verify standard interoperable checksums::

    $ fitscheck *.fits

7. Delete checksum keywords::

    $ fitscheck --checksum none --write *.fits
"""


import logging
import optparse
import sys
import textwrap
import warnings

import pyfits


log = logging.getLogger('fitscheck')


warnings.filterwarnings('error', message='Checksum verification failed')
warnings.filterwarnings('error', message='Datasum verification failed')
warnings.filterwarnings('ignore', message='Overwriting existing file')


def handle_options(args):
    if not len(args):
        args = ['-h']

    parser = optparse.OptionParser(usage=textwrap.dedent("""
        fitscheck [options] <.fits files...>

        .e.g. fitscheck example.fits

        Verifies and optionally re-writes the CHECKSUM and DATASUM keywords
        for a .fits file.
        Optionally detects and fixes FITS standard compliance problems.
        """.strip()))

    parser.add_option(
        '-k', '--checksum', dest='checksum_kind',
        type='choice', choices=['standard', 'nonstandard', 'either', 'none'],
        help='Choose FITS checksum mode or none.  Defaults standard.',
        default='standard', metavar='[standard | nonstandard | either | none]')

    parser.add_option(
        '-w', '--write', dest='write_file',
        help='Write out file checksums and/or FITS compliance fixes.',
        default=False, action='store_true')

    parser.add_option(
        '-f', '--force', dest='force',
        help='Do file update even if original checksum was bad.',
        default=False, action='store_true')

    parser.add_option(
        '-c', '--compliance', dest='compliance',
        help='Do FITS compliance checking; fix if possible.',
        default=False, action='store_true')

    parser.add_option(
        '-i', '--ignore-missing', dest='ignore_missing',
        help='Ignore missing checksums.',
        default=False, action='store_true')

    parser.add_option(
        '-v', '--verbose', dest='verbose', help='Generate extra output.',
        default=False, action='store_true')

    global OPTIONS
    OPTIONS, fits_files = parser.parse_args(args)

    if OPTIONS.checksum_kind == 'none':
        OPTIONS.checksum_kind = False

    return fits_files


def setup_logging():
    if OPTIONS.verbose:
        log.setLevel(logging.INFO)
    else:
        log.setLevel(logging.WARNING)

    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter('%(message)s'))
    log.addHandler(handler)


def verify_checksums(filename):
    """
    Prints a message if any HDU in `filename` has a bad checksum or datasum.
    """

    errors = 0
    try:
        hdulist = pyfits.open(filename, checksum=OPTIONS.checksum_kind)
    except UserWarning, w:
        remainder = '.. ' + ' '.join(str(w).split(' ')[1:]).strip()
        # if "Checksum" in str(w) or "Datasum" in str(w):
        log.warn('BAD %r %s' % (filename, remainder))
        return 1
    if not OPTIONS.ignore_missing:
        for i, hdu in enumerate(hdulist):
            if not hdu._checksum:
                log.warn('MISSING %r .. Checksum not found in HDU #%d' %
                         (filename, i))
                return 1
            if not hdu._datasum:
                log.warn('MISSING %r .. Datasum not found in HDU #%d' %
                         (filename, i))
                return 1
    if not errors:
        log.info('OK %r' % filename)
    return errors


def verify_compliance(filename):
    """Check for FITS standard compliance."""

    hdulist = pyfits.open(filename)
    try:
        hdulist.verify('exception')
    except pyfits.VerifyError, e:
        log.warn('NONCOMPLIANT %r .. %s' %
                 (filename), str(e).replace('\n',' '))
        return 1
    return 0


def update(filename):
    """
    Sets the ``CHECKSUM`` and ``DATASUM`` keywords for each HDU of `filename`.

    Also updates fixes standards violations if possible and requested.
    """

    hdulist = pyfits.open(filename)
    try:
        output_verify = 'silentfix' if OPTIONS.compliance else 'ignore'
        hdulist.writeto(filename, checksum=OPTIONS.checksum_kind, clobber=True,
                        output_verify=output_verify)
    except pyfits.VerifyError:
        pass # unfixable errors already noted during verification phase
    finally:
        hdulist.close()


def process_file(filename):
    """
    Handle a single .fits file,  returning the count of checksum and compliance
    errors.
    """

    try:
        checksum_errors = verify_checksums(filename)
        if OPTIONS.compliance:
            compliance_errors = verify_compliance(filename)
        else:
            compliance_errors = 0
        if OPTIONS.write_file and checksum_errors == 0 or OPTIONS.force:
            update(filename)
        return checksum_errors + compliance_errors
    except Exception, e:
        log.error('EXCEPTION %r .. %s' % (filename, e))
        return 1


def main():
    """
    Processes command line parameters into options and files,  then checks
    or update FITS DATASUM and CHECKSUM keywords for the specified files.
    """

    errors = 0
    fits_files = handle_options(sys.argv[1:])
    setup_logging()
    for filename in fits_files:
        errors += process_file(filename)
    if errors:
        log.warn('%d errors' % errors)
    return int(bool(errors))

